@@ -109,7 +109,7 @@ class SSHClient(BaseSSHClient):
         self._keepalive_greenlet = None
         self._proxy_client = None
         self.host = host
-        self.port = port
+        self.port = port if port is not None else 22
         if proxy_host is not None:
             _port = port if proxy_port is None else proxy_port
             _pkey = pkey if proxy_pkey is None else proxy_pkey
@@ -132,6 +132,9 @@ class SSHClient(BaseSSHClient):
             proxy_host=proxy_host, proxy_port=proxy_port,
             identity_auth=identity_auth)
 
+    def _shell(self, channel):
+        return self._eagain(channel.shell)
+
     def _connect_proxy(self, proxy_host, proxy_port, proxy_pkey,
                        user=None, password=None,
                        num_retries=DEFAULT_RETRIES,
@@ -140,6 +143,7 @@ class SSHClient(BaseSSHClient):
                        forward_ssh_agent=False,
                        keepalive_seconds=60,
                        identity_auth=True):
+        assert isinstance(self.port, int)
         try:
             self._proxy_client = SSHClient(
                 proxy_host, port=proxy_port, pkey=proxy_pkey,
@@ -251,15 +255,9 @@ class SSHClient(BaseSSHClient):
     def open_session(self):
         """Open new channel from session"""
         try:
-            chan = self.session.open_session()
+            chan = self._eagain(self.session.open_session)
         except Exception as ex:
             raise SessionError(ex)
-        while chan == LIBSSH2_ERROR_EAGAIN:
-            self.poll()
-            try:
-                chan = self.session.open_session()
-            except Exception as ex:
-                raise SessionError(ex)
         # Multiple forward requests result in ChannelRequestDenied
         # errors, flag is used to avoid this.
         if self.forward_ssh_agent and not self._forward_requested:
@@ -393,6 +391,9 @@ class SSHClient(BaseSSHClient):
         :type remote_file: str
         :param recurse: Whether or not to descend into directories recursively.
         :type recurse: bool
+        :param sftp: SFTP channel to use instead of creating a
+          new one.
+        :type sftp: :py:class:`ssh2.sftp.SFTP`
 
         :raises: :py:class:`ValueError` when a directory is supplied to
           ``local_file`` and ``recurse`` is not set
@@ -480,6 +481,9 @@ class SSHClient(BaseSSHClient):
         :type recurse: bool
         :param encoding: Encoding to use for file paths.
         :type encoding: str
+        :param sftp: SFTP channel to use instead of creating a
+          new one.
+        :type sftp: :py:class:`ssh2.sftp.SFTP`
 
         :raises: :py:class:`ValueError` when a directory is supplied to
           ``local_file`` and ``recurse`` is not set
@@ -726,6 +730,7 @@ class SSHClient(BaseSSHClient):
         Blocks current greenlet only if socket has pending read or write operations
         in the appropriate direction.
         """
+        timeout = self.timeout if timeout is None else timeout
         directions = self.session.block_directions()
         if directions == 0:
             return
@@ -748,3 +753,6 @@ class SSHClient(BaseSSHClient):
             total_written += bytes_written
             if rc == LIBSSH2_ERROR_EAGAIN:
                 self.poll(timeout=timeout)
+
+    def _eagain_write(self, write_func, data, timeout=None):
+        return self.eagain_write(write_func, data, timeout=timeout)
